#include "queue.h"

class PieceList;
class InputSpec;
class DirectionList;

// A Configuration represents an arrangement of pieces within the box.

class Configuration
{
    PieceList       *pieces;                        // the pieces in the configuration
    bool            usesGoal;                       // is the goal piece among them?
    
    int             distance;                       // these 2 fields are used for BFS, with the configuration treated as
    Configuration   *predecessor;                   // a vertex in the graph of reachable configurations.  See CLRS, p. 532.

    int             movedPiece;                     // this is additional predecessor information - how we got here from the 
    DirectionList   *movedShifts;                   // predecessor configuration.  Used by printSolution.

    static Queue    queue;                          // The BFS queue
    static  ConfigurationList *seen;                // Configurations we've already examined in the current search

public:
    
    static InputSpec    *spec;                      // the input specifications
    static int      maxDistance;                    // maximum required solution length found so far.
    
                    Configuration();
                    Configuration(Configuration &config);
                    ~Configuration();
                    
    void            puzzleSearch(int x, int y);
    void            solve();
    ConfigurationList   *findMoves(int pieceNum);
    bool            solved();
    
    void            printSolution();
    void            printSolution2();
    void            prettyPrint();
    
        
    bool            placePiece(int shapeIndex, int x, int y);
    void            removePiece();
    bool            occupied(int x, int y);
    
    bool            sameAs(Configuration *config);
};


// A ConfigurationList is a linked list of configuration pointers.  It also has the
// ability to search for a given configuration.

struct ConfigurationList
{
    Configuration       *config;
    ConfigurationList   *next;

                        ConfigurationList(Configuration *, ConfigurationList *);
                        
    void                deleteContents();
                        
    bool                isPresent(Configuration *match);
};


// A DirectionList is a linked list of directions.
// Used to record piece shift sequences that make up a move.

struct DirectionList
{
    int             dir;
    DirectionList   *next;
    
                    DirectionList(int, DirectionList *);
                    DirectionList(DirectionList *);
    void            deleteContents();
    
    void            print();
};
